import enum
import sys
import tempfile
from typing import Annotated
from typing import Literal
from typing import Optional
from typing import TypeVar

import gmsh
import numpy as np
import numpy.typing as npt
from compas.datastructures import Mesh
from compas.geometry import Brep
from compas.geometry import Point
from compas.geometry import Polyhedron
from compas.geometry import Polyline
from compas.geometry import transform_points_numpy
from compas.itertools import linspace
from compas.tolerance import TOL

from compas_gmsh.options import MeshOptions
from compas_gmsh.options import OptimizationAlgorithm

DType = TypeVar("DType", bound=np.generic)

Array3 = Annotated[npt.NDArray[DType], Literal[3]]
ArrayNx3 = Annotated[npt.NDArray[DType], Literal["N", 3]]

# from compas_gmsh.options import RecombinationAlgorithm


class MeshElementType(enum.Enum):
    Point = 15
    Line = 1
    Triangle = 2
    Quadrangle = 3
    Tetrahedron = 4
    Hexahedron = 5
    Prism = 6
    Pyramid = 7


class Model:
    """
    Base model for mesh generation.

    Parameters
    ----------
    name : str, optional
        The name of the model.
    verbose, bool, optional
        Flag indicating if output should be printed to the terminal.
    options.mesh : :class:`MeshOptions`
        The meshing options.

    """

    # gmsh.option.set_number('Mesh.MeshSizeFromCurvatureIsotropic', 0)
    # gmsh.option.set_number('Mesh.OptimizeNetgen', 0)
    # gmsh.option.set_number('Mesh.QuadsRemeshingBoldness', 0.66)
    # gmsh.option.set_number('Mesh.RecombineOptimizeTopology', 5)
    # gmsh.option.set_number('Mesh.RefineSteps', 10)
    # gmsh.option.set_number('Mesh.Smoothing', 1)
    # gmsh.option.set_number('Mesh.SmoothNormals', 0)
    # gmsh.option.set_number('Mesh.SmoothRatio', 1.8)
    # gmsh.option.set_number('Mesh.SubdivisionAlgorithm', 0)
    # gmsh.option.set_number('Mesh.ToleranceEdgeLength', 0)
    # gmsh.option.set_number('Mesh.ToleranceInitialDelaunay', 1e-12)

    class options:
        mesh = MeshOptions()

    def __init__(self, name: Optional[str] = None, verbose: bool = False) -> None:
        gmsh.initialize(sys.argv)
        gmsh.model.add(name or f"{self.__class__.__name__}")
        self._verbose = False
        self.verbose = verbose
        self.model = gmsh.model
        self.mesh = gmsh.model.mesh
        self.occ = gmsh.model.occ

    # def __del__(self):
    #     gmsh.finalize()

    def destroy(self):
        gmsh.finalize()

    @property
    def verbose(self) -> bool:
        return self._verbose

    @verbose.setter
    def verbose(self, value: bool) -> None:
        self._verbose = bool(value)
        gmsh.option.set_number("General.Terminal", int(value))

    def synchronize(self):
        """Syncronize the geometry with the underlying OCC model.

        Returns
        -------
        None

        """
        self.occ.synchronize()

    def heal(self) -> None:
        """
        Heal the underlying OCC model.

        Returns
        -------
        None

        """
        self.occ.heal_shapes()

    # ==============================================================================
    # Constructors
    # ==============================================================================

    @classmethod
    def from_step(cls, filepath: str) -> "Model":
        """Construct a model from the data contained in a STEP file.

        Parameters
        ----------
        filepath : pathlib.Path | str
            The path to the step file.

        Returns
        -------
        :class:`Model`

        """
        model = cls()
        gmsh.open(filepath)
        return model

    @classmethod
    def from_brep(cls, brep: Brep) -> "Model":
        """Construct a model from a b-rep geometry.

        Parameters
        ----------
        brep : :class:`Brep`
            The b-rep geometry.

        Returns
        -------
        :class:`Model`

        """
        fd, filepath = tempfile.mkstemp(suffix=".brep")
        brep.to_brep(filepath)  # type: ignore
        model = cls()
        gmsh.open(filepath)
        return model

    # ==============================================================================
    # Model entities
    # ==============================================================================

    @property
    def points(self) -> list[int]:
        return [dimtag[1] for dimtag in self.model.get_entities(0)]

    @property
    def curves(self) -> list[int]:
        return [dimtag[1] for dimtag in self.model.get_entities(1)]

    @property
    def surfaces(self) -> list[int]:
        return [dimtag[1] for dimtag in self.model.get_entities(2)]

    @property
    def volumes(self) -> list[int]:
        return [dimtag[1] for dimtag in self.model.get_entities(3)]

    # =============================================================================
    # Model boundaries
    # =============================================================================

    def surface_curves(self, tag: int) -> list[int]:
        """Get the boundary curves of a surface in the model.

        Parameters
        ----------
        tag : int
            The identifier of the surface.

        Returns
        -------
        list[int]
            The identifiers of the curves of the surface.

        """
        return [dimtag[1] for dimtag in self.model.get_boundary([(2, tag)])]

    def volume_surfaces(self, tag: int) -> list[int]:
        """Get the surfaces of a volume in the model.

        Parameters
        ----------
        tag : int
            The identifier of the volume.

        Returns
        -------
        list[int]
            The identifiers of the surfaces of the volume.

        """
        return [dimtag[1] for dimtag in self.model.get_boundary([(3, tag)])]

    def volume_curves(self, tag: int) -> list[int]:
        """Get the curves of a volume in the model.

        Parameters
        ----------
        tag : int
            The identifier of the volume.

        Returns
        -------
        list[int]
            The identifiers of the curves of the volume.

        """
        surfaces = self.volume_surfaces(tag)
        curves = []
        for surface in surfaces:
            temp = self.surface_curves(surface)
            curves += [tag for tag in temp if tag > 0]
        return curves

    def volume_points(self, tag: int) -> list[int]:
        """Get the points of a volume in the model.

        Parameters
        ----------
        tag : int
            The identifier of the volume.

        Returns
        -------
        list[int]
            The identifiers of the points of the volume.

        """
        dimtags = self.model.get_boundary([(3, tag)], recursive=True)
        return [dimtag[1] for dimtag in dimtags if dimtag[0] == 0]

    # =============================================================================
    # Model geometry representations
    # =============================================================================

    def curve_domain(self, tag: int) -> tuple[float, float]:
        """Get the domain of a curve in the model.

        Parameters
        ----------
        tag : int
            The identifier of the curve.

        Returns
        -------
        tuple[float, float]
            The domain of the curve.

        """
        u, v = self.model.get_parametrization_bounds(1, tag)
        return u[0], v[0]

    def surface_domain(self, tag: int) -> tuple[tuple[float, float], tuple[float, float]]:
        """Get the domain of a surface in the model.

        Parameters
        ----------
        tag : int
            The identifier of the surface.

        Returns
        -------
        tuple[tuple[float, float], tuple[float, float]]
            The domain of the surface.

        """
        dmin, dmax = self.model.get_parametrization_bounds(2, tag)
        umin, vmin = dmin
        umax, vmax = dmax
        return (umin, umax), (vmin, vmax)

    def point_coordinates(self, tag: int) -> Array3[np.float64]:
        """Get the coordinates of a point in the model.

        Parameters
        ----------
        tag : int
            The identifier of the point.

        Returns
        -------
        list[float]
            The coordinates of the point.

        """
        return self.model.get_value(0, tag, [None])  # type: ignore

    def curve_coordinates_at(self, tag: int, u: float) -> Array3[np.float64]:
        """Get the coordinates of a point on a curve in the model.

        Parameters
        ----------
        tag : int
            The identifier of the curve.
        u : float
            The parameter of the curve.

        Returns
        -------
        list[float]
            The coordinates of the point on the curve.

        """
        return self.model.get_value(1, tag, [u])  # type: ignore

    def surface_coordinates_at(self, tag: int, u: float, v: float) -> Array3[np.float64]:
        """Get the coordinates of a point on a surface in the model.

        Parameters
        ----------
        tag : int
            The identifier of the surface.
        u : float
            The first parameter of the surface.
        v : float
            The second parameter of the surface.

        Returns
        -------
        list[float]
            The coordinates of the point on the surface.

        """
        return self.model.get_value(2, tag, [u, v])  # type: ignore

    # =============================================================================
    # Model geometry conversions
    # =============================================================================

    def point_to_point(self, tag: int) -> Point:
        """Get the coordinates of a point in the model.

        Parameters
        ----------
        tag : int
            The identifier of the point.

        Returns
        -------
        :class:`Point`
            The point.

        """
        return Point(*self.point_coordinates(tag))

    def curve_to_polyline(self, tag: int, n: int = 100) -> Polyline:
        """Get the coordinates of a point on a curve in the model.

        Parameters
        ----------
        tag : int
            The identifier of the curve.
        n : int, optional
            The number of points on the curve.

        Returns
        -------
        :class:`Polyline`
            The polyline that is a discretisation of the curve.

        """
        u, v = self.curve_domain(tag)
        points = [self.curve_coordinates_at(tag, t) for t in linspace(u, v, n)]
        return Polyline(points)

    # =============================================================================
    # Model exploration
    # =============================================================================

    def find_points_at_xyz(self, xyz: list[float], tolerance=None) -> list[int]:
        """Find the model points at or close to a spatial location.

        Parameters
        ----------
        xyz : list of float
            The XYZ coordinates of the location.

        Returns
        -------
        list of int
            The point identifiers.

        """
        points = []
        key = TOL.geometric_key(xyz)
        for point in self.points:
            test = TOL.geometric_key(self.point_coordinates(point))
            if test == key:
                points.append(point)
        return points

    def find_points_at_xy(self, xyz: list[float], tolerance=None) -> list[int]:
        """Find the model points at or close to a spatial location.

        Parameters
        ----------
        xyz : list of float
            The XY(Z) coordinates of the location.

        Returns
        -------
        list of int
            The point identifiers.

        """
        points = []
        key = TOL.geometric_key_xy(xyz[:2])
        for point in self.points:
            test = TOL.geometric_key_xy(self.point_coordinates(point)[:2])
            if test == key:
                points.append(point)
        return points

    # ==============================================================================
    # Meshing
    # ==============================================================================

    def mesh_targetlength_at_point(self, tag: int, target: float) -> None:
        """
        Set the target length at a particular mesh point.

        Parameters
        ----------
        tag : int
            The point identifier.
        target : float
            The target length value.

        Returns
        -------
        None

        """
        self.occ.mesh.set_size([(0, tag)], target)

    def generate_mesh(self, dim: int = 2) -> None:
        """
        Generate a mesh of the current model.

        Parameters
        ----------
        dim : int, optional
            The dimension of the mesh.

        Returns
        -------
        None
            The mesh is stored in the model for further refinement and optimisation.
            To retrieve the generated mesh, use :meth:`mesh_to_compas`, :meth:`mesh_to_openmesh`, or :meth:`mesh_to_tets`.

        Notes
        -----
        The geometry is automatically synchronised with the underlying OCC model.
        Therefore, there is no need to call :meth:`synchronize` before generating the mesh.
        To influence the meshing process, use the options of the model (:attr:`options.mesh`).

        """
        self.occ.synchronize()
        self.mesh.generate(dim)

    def refine_mesh(self) -> None:
        """
        Refine the model mesh by uniformly splitting the edges.

        Returns
        -------
        None
            Refinement is applied to the internally stored mesh.

        """
        self.mesh.refine()

    def optimize_mesh(
        self,
        algo: OptimizationAlgorithm = OptimizationAlgorithm.Default,
        niter: int = 1,
    ) -> None:
        """
        Optimize the model mesh using the specified method.

        Parameters
        ----------
        algo : OptimizationAlgorithm, optional
            The optimization algorithm to use.
        niter : int, optional
            The number of iterations.

        Returns
        -------
        None
            Optimisation is applied to the internally stored mesh.

        """
        self.mesh.optimize(algo.value, niter=niter)

    def recombine_mesh(self) -> None:
        """
        Recombine the mesh into quadrilateral faces.

        Returns
        -------
        None
            Recombination is applied to the internally stored mesh.

        """
        self.mesh.recombine()

    # =============================================================================
    # Mesh elements
    # =============================================================================

    def curve_vertices(self, tag: int) -> dict[int, Array3[np.float64]]:
        """Get the vertices of a curve.

        Parameters
        ----------
        tag : int
            The identifier of the curve.

        Returns
        -------
        dict[int, list[float]]
            The vertices of the curve.

        """
        node_coords: ArrayNx3[np.float64]
        node_tags, node_coords, _ = self.mesh.get_nodes_by_element_type(1, tag)  # type: ignore
        return dict(zip(node_tags, node_coords.reshape((-1, 3), order="C")))

    def surface_vertices(self, tag: int) -> dict[int, Array3[np.float64]]:
        """Get the vertices of a surface.

        Parameters
        ----------
        tag : int
            The identifier of the surface.

        Returns
        -------
        list[list[float]]
            The vertices of the surface.

        """
        node_coords: ArrayNx3[np.float64]
        node_tags, node_coords, _ = self.mesh.get_nodes_by_element_type(2, tag)  # type: ignore
        return dict(zip(node_tags, node_coords.reshape((-1, 3), order="C")))

    def surface_faces(self, tag: int) -> list[list[int]]:
        """Get the faces of a surface.

        Parameters
        ----------
        tag : int
            The identifier of the surface.

        Returns
        -------
        list[list[int]]
            The faces of the surface.

        """
        faces = []
        elements = self.mesh.get_elements(2, tag)
        for element_type, element_tags, node_tags in zip(*elements):
            element_props = self.mesh.get_element_properties(element_type)
            number_of_nodes = element_props[3]
            for i, element_tag in enumerate(element_tags):
                nodes = node_tags[i * number_of_nodes : (i + 1) * number_of_nodes]
                faces.append(nodes)
        return faces

    def surface_mesh(self, tag: int) -> Mesh:
        """Get the mesh of a surface entity.

        Parameters
        ----------
        tag : int
            The identifier of the surface.

        Returns
        -------
        :class:`Mesh`
            The mesh of the surface.

        """
        vertices = self.surface_vertices(tag)
        faces = self.surface_faces(tag)
        mesh = Mesh.from_vertices_and_faces(vertices, faces)
        mesh.remove_unused_vertices()
        return mesh

    def volume_vertices(self, tag: int) -> dict[int, Array3[np.float64]]:
        """Get the vertices of a volume.

        Parameters
        ----------
        tag : int
            The identifier of the volume.

        Returns
        -------
        dict[int, list[float]]
            The vertices of the volume.

        """
        vertices = {}
        _, downward = self.model.get_adjacencies(3, tag)
        node_coords: ArrayNx3[np.float64]
        for tag in downward:
            node_tags, node_coords, _ = self.mesh.get_nodes_by_element_type(2, tag)  # type: ignore
            vertices.update(dict(zip(node_tags, node_coords.reshape((-1, 3), order="C"))))
        return vertices

    def volume_faces(self, tag: int) -> list[list[int]]:
        """Get the faces of a volume.

        Parameters
        ----------
        tag : int
            The identifier of the volume.

        Returns
        -------
        list[list[int]]
            The faces of the volume.

        """
        _, downward = self.model.get_adjacencies(3, tag)
        faces = []
        for tag in downward:
            elements = self.mesh.get_elements(2, tag)
            for element_type, element_tags, node_tags in zip(*elements):
                element_props = self.mesh.get_element_properties(element_type)
                number_of_nodes = element_props[3]
                for i, element_tag in enumerate(element_tags):
                    faces.append(node_tags[i * number_of_nodes : (i + 1) * number_of_nodes])
        return faces

    def volume_mesh(self, tag: int) -> Mesh:
        """Get the mesh of a volume entity.

        Parameters
        ----------
        tag : int
            The identifier of the volume.

        Returns
        -------
        :class:`Mesh`
            The mesh of the volume.

        """
        vertices = self.volume_vertices(tag)
        faces = self.volume_faces(tag)
        mesh = Mesh.from_vertices_and_faces(vertices, faces)
        mesh.remove_unused_vertices()
        return mesh

    def volume_tets(self, tag: int) -> list[Polyhedron]:
        """Get the tetrahedra of a volume entity.

        Parameters
        ----------
        tag : int
            The identifier of the volume.

        Returns
        -------
        list[:class:`Polyhedron`]
            The tetrahedra of the volume.

        """
        raise NotImplementedError

    # ==============================================================================
    # Export
    # ==============================================================================

    def mesh_node_xyz(self):
        """Get the coordinates of the mesh nodes.

        Returns
        -------
        dict
            A dictionary mapping node tags to node coordinates.

        """
        node_coords: ArrayNx3[np.float64]
        node_tags, node_coords, _ = self.mesh.get_nodes()  # type: ignore
        return dict(zip(node_tags, node_coords.reshape((-1, 3), order="C")))

    def mesh_to_vertices_and_faces(
        self,
    ) -> tuple[dict[int, Array3[np.float64]], list[list[int]]]:
        """
        Convert the model mesh to a COMPAS mesh data structure.

        Returns
        -------
        tuple[dict, list]
            A tuple containing the vertices and faces of the mesh.

        """
        vertices = self.mesh_node_xyz()

        # element_types is a list of element types
        # element_tags is a list containing a list of tags for each element type
        # node_tags is a list containing a flattened list of node tags for each element type
        elements = self.mesh.get_elements()
        faces = []
        for element_type, element_tags, node_tags in zip(*elements):
            number_of_nodes = self.mesh.get_element_properties(element_type)[3]
            if element_type in (2, 3):
                for i, element_tag in enumerate(element_tags):
                    face = node_tags[i * number_of_nodes : i * number_of_nodes + number_of_nodes]
                    faces.append(face)

        return vertices, faces

    def mesh_to_triangles(self) -> list[list[Point]]:
        """Convert the model mesh to a list of triangles.

        Returns
        -------
        list[(Point, Point, Point)]
            A list of triangles.

        """
        # set the element type to triangles
        element_type = 2
        # get all triangle nodes
        coords: ArrayNx3[np.float64]
        tags, coords, _ = self.mesh.get_nodes_by_element_type(element_type, returnParametricCoord=False)  # type: ignore
        node_xyz = dict(zip(tags, coords.reshape((-1, 3), order="C")))
        # get properties of triangles
        element_props = self.mesh.get_element_properties(element_type)
        # get number of nodes per triangle
        number_of_nodes = element_props[3]
        # get all triangles
        # element_tags is a list of tags
        # node_tags is a flattened list of node tags
        element_tags, node_tags = self.mesh.get_elements_by_type(element_type)
        # reshape the node_tags array to a 2D array
        node_tags = node_tags.reshape((-1, number_of_nodes), order="C")  # type: ignore
        # every row in node_tags is a triangle face
        triangles = [[Point(*node_xyz[node]) for node in face] for face in node_tags]
        return triangles

    def mesh_to_quads(self) -> list[list[Point]]:
        """Convert the model mesh to a list of quads.

        Returns
        -------
        list[(Point, Point, Point, Point)]
            A list of quads.

        """
        # set the element type to quads
        element_type = 3
        # get all quad nodes
        coords: ArrayNx3[np.float64]
        tags, coords, _ = self.mesh.get_nodes_by_element_type(element_type, returnParametricCoord=False)  # type: ignore
        node_xyz = dict(zip(tags, coords.reshape((-1, 3), order="C")))
        # get properties of quads
        element_props = self.mesh.get_element_properties(element_type)
        # get number of nodes per quad
        number_of_nodes = element_props[3]
        # get all quads
        # element_tags is a list of tags
        # node_tags is a flattened list of node tags
        element_tags, node_tags = self.mesh.get_elements_by_type(element_type)
        # reshape the node_tags array to a 2D array
        node_tags = node_tags.reshape((-1, number_of_nodes), order="C")  # type: ignore
        # every row in node_tags is a quad face
        quads = [[Point(*node_xyz[node]) for node in face] for face in node_tags]
        return quads

    def mesh_to_tets(self) -> list[Polyhedron]:
        """
        Convert the model mesh to a COMPAS mesh data structure.

        Returns
        -------
        list[:class:`Polyhedron`]
            A list of COMPAS polyhedra.

        """
        # set the element type to tetrahedra
        element_type = 4
        # get all tetrahedra nodes
        coords: ArrayNx3[np.float64]
        tags, coords, _ = self.mesh.get_nodes_by_element_type(element_type, returnParametricCoord=False)  # type: ignore
        # make a node coordinate map
        node_xyz = dict(zip(tags, coords.reshape((-1, 3), order="C")))
        # get properties of tetrahedra
        element_props = self.mesh.get_element_properties(element_type)
        # get number of nodes per tetrahedron
        number_of_nodes = element_props[3]
        # get all tetrahedra
        element_tags, node_tags = self.mesh.get_elements_by_type(element_type)
        node_tags = node_tags.reshape((-1, number_of_nodes), order="C")  # type: ignore
        # construct a polyhedron for each tetrahedron
        tets = []
        for nodes in node_tags:
            vertices = [node_xyz[node] for node in nodes]
            faces = [[0, 1, 2], [0, 2, 3], [1, 3, 2], [0, 3, 1]]
            tets.append(Polyhedron(vertices, faces))
        return tets

    def mesh_to_vertices_faces_edges(
        self,
        transformation=None,
    ) -> tuple[list[list[float]], list[list[int]], list[tuple[int, int]]]:
        # set the element type to tetrahedra
        element_type = 4
        # get all tetrahedra nodes
        coords: ArrayNx3[np.float64]
        tags, coords, _ = self.mesh.get_nodes_by_element_type(element_type, returnParametricCoord=False)  # type: ignore
        # make a node coordinate map
        node_index = {tag: index for index, tag in enumerate(tags)}
        # get properties of tetrahedra
        element_props = self.mesh.get_element_properties(element_type)
        # get number of nodes per tetrahedron
        number_of_nodes = element_props[3]
        # get all tetrahedra
        element_tags, node_tags = self.mesh.get_elements_by_type(element_type)
        node_tags = node_tags.reshape((-1, number_of_nodes), order="C")  # type: ignore
        # construct the vertices faces and edges
        xyz = coords.reshape((-1, 3), order="C")
        if transformation:
            xyz = transform_points_numpy(xyz, transformation)
        vertices = xyz.tolist()
        faces = []
        edges = []
        for nodes in node_tags:
            faces.append([node_index[nodes[n]] for n in [0, 1, 2]])
            faces.append([node_index[nodes[n]] for n in [0, 2, 3]])
            faces.append([node_index[nodes[n]] for n in [1, 3, 2]])
            faces.append([node_index[nodes[n]] for n in [0, 3, 1]])
            edges.append((node_index[nodes[0]], node_index[nodes[1]]))
            edges.append((node_index[nodes[1]], node_index[nodes[2]]))
            edges.append((node_index[nodes[2]], node_index[nodes[0]]))
            edges.append((node_index[nodes[0]], node_index[nodes[3]]))
            edges.append((node_index[nodes[1]], node_index[nodes[3]]))
            edges.append((node_index[nodes[2]], node_index[nodes[3]]))
        return vertices, faces, edges

    def mesh_to_faces(self) -> list[list[list[float]]]:
        # set the element type to tetrahedra
        element_type = 4
        # get all tetrahedra nodes
        coords: ArrayNx3[np.float64]
        tags, coords, _ = self.mesh.get_nodes_by_element_type(element_type, returnParametricCoord=False)  # type: ignore
        # make a node coordinate map
        node_xyz = dict(zip(tags, coords.reshape((-1, 3), order="C")))
        # get properties of tetrahedra
        element_props = self.mesh.get_element_properties(element_type)
        # get number of nodes per tetrahedron
        number_of_nodes = element_props[3]
        # get all tetrahedra
        element_tags, node_tags = self.mesh.get_elements_by_type(element_type)
        node_tags = node_tags.reshape((-1, number_of_nodes), order="C")  # type: ignore
        # construct a polyhedron for each tetrahedron
        tets = []
        for nodes in node_tags:
            vertices = [node_xyz[node] for node in nodes]
            faces = [
                [vertices[0], vertices[1], vertices[2]],
                [vertices[0], vertices[2], vertices[3]],
                [vertices[1], vertices[3], vertices[2]],
                [vertices[0], vertices[3], vertices[1]],
            ]
            tets.append(faces)
        return tets

    def mesh_to_compas(self) -> Mesh:
        """
        Convert the model mesh to a COMPAS mesh data structure.

        Returns
        -------
        :class:`Mesh`
            A COMPAS mesh.

        """
        vertices, faces = self.mesh_to_vertices_and_faces()
        return Mesh.from_vertices_and_faces(vertices, faces)

    def mesh_to_openmesh(self) -> None:  # type: ignore
        """
        Convert the model mesh to a COMPAS mesh data structure.

        Returns
        -------
        openmesh.TriMesh or openmesh.PolyMesh
            An OpenMesh mesh.

        """
        try:
            import openmesh  # type: ignore
        except ImportError:
            print("OpenMesh is not installed. Install using `pip install openmesh`.")
            raise

        vertices, faces = self.mesh_to_vertices_and_faces()

        if len(faces[0]) == 3:
            mesh = openmesh.TriMesh()
        elif len(faces[0]) == 4:
            mesh = openmesh.PolyMesh()
        else:
            raise NotImplementedError

        vertex_index = {}
        for vertex in vertices:
            index = mesh.add_vertex(vertices[vertex])
            vertex_index[vertex] = index
        for face in faces:
            mesh.add_face(*[vertex_index[vertex] for vertex in face])

        return mesh



from typing import List
from typing import Tuple

from compas.geometry import Box
from compas.geometry import Cylinder
from compas.geometry import Sphere

from .model import Model


class ShapeModel(Model):
    """
    Model for shape combinations.
    """

    def add_cylinder(self, cylinder: Cylinder) -> Tuple[int, int]:
        """Add a cylinder to the model."""
        H = cylinder.height
        R = cylinder.radius
        base = cylinder.plane.point
        normal = cylinder.plane.normal
        start = base + normal.scaled(-0.5 * H)
        vector = normal.scaled(H)
        x0, y0, z0 = start
        dx, dy, dz = vector
        tag = self.occ.add_cylinder(x0, y0, z0, dx, dy, dz, R)
        return 3, tag

    def add_sphere(self, sphere: Sphere) -> Tuple[int, int]:
        """Add a sphere to the model."""
        x, y, z = sphere.frame.point
        R = sphere.radius
        tag = self.occ.add_sphere(x, y, z, R)
        return 3, tag

    def add_box(self, box: Box) -> Tuple[int, int]:
        """Add a box to the model."""
        x0, y0, z0 = box.frame.point
        x = x0 - 0.5 * box.xsize
        y = y0 - 0.5 * box.ysize
        z = z0 - 0.5 * box.zsize
        tag = self.occ.add_box(x, y, z, box.xsize, box.ysize, box.zsize)
        return 3, tag

    def boolean_intersection(
        self,
        A: List[Tuple[int, int]],
        B: List[Tuple[int, int]],
        remove_objects: bool = True,
        remove_tools: bool = True,
    ) -> List[Tuple[int, int]]:
        """
        Boolean intersection of two sets of shapes.

        Parameters
        ----------
        A : list of tuple
            The *dimtags* of the *object* shapes.
        B : tuple
            The *dimtags* of the *tool* shapes.

        Returns
        -------
        list of tuple
            The dimtags of the resulting shapes.

        Notes
        -----
        The *objects* are the shapes to which the boolean operation should be applied.
        The *tools* are the shapes used to perform the operation.

        """
        result = self.occ.intersect(A, B, removeObject=remove_objects, removeTool=remove_tools)
        dimtags = result[0]
        return dimtags

    def boolean_union(
        self,
        A: List[Tuple[int, int]],
        B: List[Tuple[int, int]],
        remove_objects: bool = True,
        remove_tools: bool = True,
    ) -> List[Tuple[int, int]]:
        """
        Boolean union of two sets of shapes.

        Parameters
        ----------
        A : list of tuple
            The *dimtags* of the *object* shapes.
        B : tuple
            The *dimtags* of the *tool* shapes.

        Returns
        -------
        list of tuple
            The dimtags of the resulting shapes.

        Notes
        -----
        The *objects* are the shapes to which the boolean operation should be applied.
        The *tools* are the shapes used to perform the operation.

        """
        result = self.occ.fuse(A, B, removeObject=remove_objects, removeTool=remove_tools)
        dimtags = result[0]
        return dimtags

    def boolean_difference(
        self,
        A: List[Tuple[int, int]],
        B: List[Tuple[int, int]],
        remove_objects: bool = True,
        remove_tools: bool = True,
    ) -> List[Tuple[int, int]]:
        """
        Boolean difference of two sets of shapes.

        Parameters
        ----------
        A : list of tuple
            The *dimtags* of the *object* shapes.
        B : tuple
            The *dimtags* of the *tool* shapes.

        Returns
        -------
        list of tuple
            The dimtags of the resulting shapes.

        Notes
        -----
        The *objects* are the shapes to which the boolean operation should be applied.
        The *tools* are the shapes used to perform the operation.

        """
        result = self.occ.cut(A, B, removeObject=remove_objects, removeTool=remove_tools)
        dimtags = result[0]
        return dimtags

    def boolean_fragment(
        self,
        A: List[Tuple[int, int]],
        B: List[Tuple[int, int]],
        remove_objects: bool = True,
        remove_tools: bool = True,
    ) -> List[Tuple[int, int]]:
        """
        Boolean fragment of two sets of shapes.

        Parameters
        ----------
        A : list of tuple
            The *dimtags* of the *object* shapes.
        B : tuple
            The *dimtags* of the *tool* shapes.

        Returns
        -------
        list of tuple
            The dimtags of the resulting shapes.

        Notes
        -----
        The *objects* are the shapes to which the boolean operation should be applied.
        The *tools* are the shapes used to perform the operation.

        """
        result = self.occ.fragment(A, B, removeObject=remove_objects, removeTool=remove_tools)
        dimtags = result[0]
        return dimtags



from typing import Dict
from typing import Optional
from typing import Union

from compas.datastructures import Mesh

from .model import Model


class MeshModel(Model):
    """Model for mesh (re)meshing."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.vertex_tag = {}

    @classmethod
    def from_mesh(
        cls: "MeshModel",
        mesh: Mesh,
        name: str = "Mesh",
        targetlength: Optional[Union[float, Dict]] = None,
    ) -> "MeshModel":
        """
        Create a mesh model from a mesh.

        Parameters
        ----------
        mesh : :class:`Mesh`
            A COMPAS mesh.
        name : str, optional
            The name of the model.
        target_length : float, optional
            Target value for the length of the mesh edges.

        Returns
        -------
        :class:`MeshModel`

        """
        model: MeshModel = cls(name)

        for vertex in mesh.vertices():
            point = mesh.vertex_coordinates(vertex)
            if targetlength:
                if isinstance(targetlength, dict):
                    length = targetlength.get(vertex)
                else:
                    length = targetlength
                tag = model.occ.add_point(*point, length)
            else:
                tag = model.occ.add_point(*point)
            model.vertex_tag[vertex] = tag

        for face in mesh.faces():
            loop = []
            for u, v in mesh.face_halfedges(face):
                tag = model.occ.add_line(model.vertex_tag[u], model.vertex_tag[v])
                loop.append(tag)
            tag = model.occ.add_curve_loop(loop)
            model.occ.add_surface_filling(tag)

        model.synchronize()
        return model

    def generate_mesh(self, dim: int = 2) -> None:
        """
        Generate a mesh of the current model.

        Parameters
        ----------
        dim : int, optional
            The dimension of the mesh.

        Returns
        -------
        None
            The mesh is stored in the model for further refinement and optimisation.
            To retrieve the generated mesh, use :meth:`mesh_to_compas`, :meth:`mesh_to_openmesh`, or :meth:`mesh_to_tets`.

        Notes
        -----
        The geometry is automatically synchronised with the underlying OCC model.
        Therefore, there is no need to call :meth:`synchronize` before generating the mesh.
        To influence the meshing process, use the options of the model (:attr:`options.mesh`).

        """
        self.occ.heal_shapes()
        self.occ.synchronize()
        self.mesh.generate(dim)

    def find_point_at_vertex(self, vertex: int) -> int:
        """Find the model point at a vertex of the input mesh.

        Parameters
        ----------
        vertex : int
            A vertex of the input mesh.

        Returns
        -------
        int
            The point identifier.

        """
        return self.vertex_tag[vertex]

    def mesh_targetlength_at_vertex(self, vertex: int, target: float) -> None:
        """
        Set the target length at a particular mesh vertex.

        Parameters
        ----------
        vertex : int
            The vertex identifier.
        target : float
            The target length value.

        Returns
        -------
        None

        """
        tag = self.vertex_tag[vertex]
        self.occ.mesh.set_size([(0, tag)], target)

    def read_options_from_attributes(self) -> None:
        """Read the model options from the attributes of the mesh data structure.

        Returns
        -------
        None

        """
        pass



from typing import List
from typing import Tuple

import compas.geometry
from compas.geometry import Box
from compas.geometry import Cylinder
from compas.geometry import Sphere

from .model import Model


def add_cylinder(self, cylinder: Cylinder) -> Tuple[int, int]:
    """Add a cylinder to the model."""
    H = cylinder.height
    R = cylinder.radius
    base = cylinder.plane.point
    normal = cylinder.plane.normal
    start = base + normal.scaled(-0.5 * H)
    vector = normal.scaled(H)
    x0, y0, z0 = start
    dx, dy, dz = vector
    tag = self.occ.addCylinder(x0, y0, z0, dx, dy, dz, R)
    return 3, tag


def add_sphere(self, sphere: Sphere) -> Tuple[int, int]:
    """Add a sphere to the model."""
    x, y, z = sphere.point
    R = sphere.radius
    tag = self.occ.addSphere(x, y, z, R)
    return 3, tag


def add_box(self, box: Box) -> Tuple[int, int]:
    """Add a box to the model."""
    x0, y0, z0 = box.frame.point
    x = x0 - 0.5 * box.xsize
    y = y0 - 0.5 * box.ysize
    z = z0 - 0.5 * box.zsize
    tag = self.occ.addBox(x, y, z, box.xsize, box.ysize, box.zsize)
    return 3, tag


_SHAPE_FUNC = {Box: add_box, Sphere: add_sphere, Cylinder: add_cylinder}


class CSGModel(Model):
    """Model for shape generation through Constructive Solid Geometry.

    Parameters
    ----------
    tree : dict
        The CSG tree as a dictionary mapping operations to operands.
        The operations have to be one of ``{"union", "intersection", "difference"}``.
        The operands have to be lists of shapes or lists of dicts that are CSG trees themselves.
        At every level of the tree, there can be only one operation.
    name : str
        The name of the model.

    Attributes
    ----------
    tree : dict
        The CSG tree as a dictionary mapping operations to operands.
        The operations have to be one of ``{"union", "intersection", "difference"}``.
        The operands have to be lists of shapes or lists of dicts that are CSG trees themselves.
        At every level of the tree, there can be only one operation.

    """

    def __init__(self, tree: dict, name: str, **kwargs) -> None:
        super().__init__(name, **kwargs)
        self._tree = None
        self.tree = tree

    @property
    def tree(self) -> dict:
        return self._tree

    @tree.setter
    def tree(self, tree: dict) -> None:
        def check(tree: dict) -> None:
            if len(tree) > 1:
                raise Exception("The tree can only have 1 operation per level.")
            operation = next(iter(tree))
            operands = tree[operation]
            for operand in operands:
                if isinstance(operand, dict):
                    check(operand)

        try:
            check(tree)
        except Exception:
            raise
        else:
            self._tree = tree

    def add(self, shape: compas.geometry.Shape) -> Tuple[int, int]:
        """
        Add a shape to the underlying OCC model.

        Parameters
        ----------
        shape : :class:`compas.geometry.Shape`
            A geometric shape.

        Returns
        -------
        tuple[int, int]
            A "dimtag", a dimension and a tag that together uniquely identify the shape in the OCC model.

        """
        stype = type(shape)
        if stype not in _SHAPE_FUNC:
            raise Exception(f"Shape type is not supported: {stype}")
        dimtag = _SHAPE_FUNC[type(shape)](self, shape)
        return dimtag

    def compute_tree(self) -> None:
        """
        Comute the compound shape resulting from the operations on shape primitives in the tree.

        Returns
        -------
        None

        """

        def walk(tree: dict) -> List[Tuple[int, int]]:
            operation = next(iter(tree))
            operands = tree[operation]

            for index, operand in enumerate(operands):
                if isinstance(operand, dict):
                    operands[index] = walk(operand)
                elif isinstance(operand, list):
                    operands[index] = [self.add(o) for o in operand]
                else:
                    operands[index] = self.add(operand)

            if operation == "union":
                result = self.occ.fuse(operands[:-1], operands[-1:])
                return result[0][0]

            if operation == "difference":
                result = self.occ.cut(operands[:-1], operands[-1:])
                return result[0][0]

            if operation == "intersection":
                result = self.occ.intersect(operands[:-1], operands[-1:])
                return result[0][0]

        walk(self.tree)



from enum import Enum

import gmsh


class MeshAlgorithm(Enum):
    MeshAdapt = 1
    Automatic = 2
    InitialMeshOnly = 3
    Delaunay = 5
    FrontalDelaunay = 6
    BAMG = 7
    FrontalDelaunayQuads = 8
    PackingParallelograms = 9

    @classmethod
    def __missing__(cls, value):
        return cls.MeshAdapt


class OptimizationAlgorithm(Enum):
    Default = ""
    Gmsh = "Gmsh"
    Netgen = "Netgen"
    HighOrder = "HighOrder"
    HighOrderElastic = "HighOrderElastic"
    HighOrderFastCurving = "HighOrderFastCurving"
    Laplace2D = "Laplace2D"
    Relocate2D = "Relocate2D"
    Relocate3D = "Relocate3D"

    @classmethod
    def __missing__(cls, value):
        return cls.Default


class RecombinationAlgorithm(Enum):
    Simple = 0
    Blossom = 1
    SimpleFullQuad = 2
    BlossomFullQuad = 3

    @classmethod
    def __missing__(cls, value):
        return cls.Blossom


class MeshSizeExtendFromBoundary(Enum):
    Never = 0
    SurfacesAndVolumes = 1
    SurfacesAndVolumesUseSmallestSurfaceElementEdgeLength = 2
    OnlySurfaces = -2
    OnlyVolumes = -3

    @classmethod
    def __missing__(cls, value):
        return cls.SurfacesAndVolumes


class MeshOptions:
    """Options for mesh generation in Gmsh.

    Attributes
    ----------
    algorithm : MeshAlgorithm
        The meshing algorithm to use.
    lmin : float
        The minimum characteristic edge length for meshing.
        Note that the actual minimum edge length may deviate from this value.
        The deviation is captured in an efficiency index, which tends to be around 0.8.
        Use :attr:`MeshAlgortihmFrontalDelaunay` for best results.
    lmax : float
        The maximum characteristic edge length for meshing.
        Note that the actual maximum edge length may deviate from this value.
        The deviation is captured in an efficiency index, which tends to be around 0.8.
        Use :attr:`MeshAlgortihmFrontalDelaunay` for best results.
    mesh_only_empty : bool
        Mesh only parts without existing mesh.
        Default is False.
    meshsize_extend_from_boundary : MeshSizeExtendFromBoundary
        Compute mesh size from the boundary inwards.
    meshsize_min : float
        Minimum size of mesh elements.
        Default is 0.
    meshsize_max : float
        Maximum size of mesh elements.
        Default is 1e+22.
    meshsize_from_curvature : bool
        Define mesh size based on curvature.
        Default is False.
    meshsize_from_points : bool
        Define mesh size based values assigned to points.
        Default is True.
    min_nodes_circle : float
        Minimum number of nodes for discretising a circle.
        Default is 7.
    min_nodes_curve : float
        Minimum number of nodes for discretising a curve.
        Default is 3.

    Notes
    -----
    There are three ways to specify the size of the mesh elements for a given geometry:

    1. First, if Mesh.CharacteristicLengthFromPoints is set (it is by default),
       you can simply specify desired mesh element sizes at the geometrical points of the model (with the Point command: see Points).
       The size of the mesh elements will then be computed by linearly interpolating these values on the initial mesh (see Mesh: finite element mesh generation).
       This might sometimes lead to over-refinement in some areas, so that you may have to add “dummy” geometrical entities in the model in order to get the desired element sizes.
       This method works with all the algorithms implemented in the mesh module.
       The final element sizes are of course constrained by the structured algorithms for which the element sizes are explicitly specified
       (e.g., transfinite and extruded grids: see Structured grids).
    2. Second, if Mesh.CharacteristicLengthFromCurvature is set (it is not by default), the mesh will be adapted with respect to the curvature of the geometrical entities.
    3. Finally, you can specify general mesh size "fields".

    Fields are supported by all the algorithms except those based on Netgen.
    The list of available fields with their options is given below.
    The three aforementioned methods can be used simultaneously, in which case the smallest element size is selected at any given point.
    All element sizes are further constrained by the Mesh.CharacteristicLengthMin, Mesh.CharacteristicLengthMax and Mesh.CharacteristicLengthFactor.

    """

    @property
    def algorithm(self) -> MeshAlgorithm:
        return MeshAlgorithm(gmsh.option.get_number("Mesh.Algorithm"))

    @algorithm.setter
    def algorithm(self, algo: MeshAlgorithm) -> None:
        gmsh.option.set_number("Mesh.Algorithm", algo.value)

    @property
    def lmin(self) -> float:
        gmsh.option.get_number("Mesh.CharacteristicLengthMin")

    @lmin.setter
    def lmin(self, value: float):
        gmsh.option.set_number("Mesh.CharacteristicLengthMin", value)

    @property
    def lmax(self) -> float:
        gmsh.option.get_number("Mesh.CharacteristicLengthMax")

    @lmax.setter
    def lmax(self, value: float):
        gmsh.option.set_number("Mesh.CharacteristicLengthMax", value)

    @property
    def mesh_only_empty(self) -> bool:
        return bool(gmsh.option.get_number("Mesh.MeshOnlyEmpty"))

    @mesh_only_empty.setter
    def mesh_only_empty(self, value: bool):
        gmsh.option.set_number("Mesh.MeshOnlyEmpty", int(value))

    @property
    def meshsize_extend_from_boundary(self) -> MeshSizeExtendFromBoundary:
        return MeshSizeExtendFromBoundary(gmsh.option.get_number("Mesh.MeshSizeExtendFromBoundary"))

    @meshsize_extend_from_boundary.setter
    def meshsize_extend_from_boundary(self, value: MeshSizeExtendFromBoundary):
        gmsh.option.set_number("Mesh.MeshSizeExtendFromBoundary", int(value))

    @property
    def meshsize_min(self) -> float:
        return gmsh.option.get_number("Mesh.MeshSizeMin")

    @meshsize_min.setter
    def meshsize_min(self, value: float):
        gmsh.option.set_number("Mesh.MeshSizeMin", value)

    @property
    def meshsize_max(self) -> float:
        return gmsh.option.get_number("Mesh.MeshSizeMax")

    @meshsize_max.setter
    def meshsize_max(self, value: float):
        gmsh.option.set_number("Mesh.MeshSizeMax", value)

    @property
    def meshsize_from_curvature(self) -> bool:
        return bool(gmsh.option.get_number("Mesh.MeshSizeFromCurvature"))

    @meshsize_from_curvature.setter
    def meshsize_from_curvature(self, value: bool):
        gmsh.option.set_number("Mesh.MeshSizeFromCurvature", int(value))

    @property
    def meshsize_from_points(self) -> bool:
        return bool(gmsh.option.get_number("Mesh.MeshSizeFromPoints"))

    @meshsize_from_points.setter
    def meshsize_from_points(self, value: bool):
        gmsh.option.set_number("Mesh.MeshSizeFromPoints", int(value))

    @property
    def min_nodes_circle(self) -> float:
        return int(gmsh.option.get_number("Mesh.MinimumCircleNodes"))

    @min_nodes_circle.setter
    def min_nodes_circle(self, value: float):
        gmsh.option.set_number("Mesh.MinimumCircleNodes", value)

    @property
    def min_nodes_curve(self) -> float:
        return int(gmsh.option.get_number("Mesh.MinimumCurveNodes"))

    @min_nodes_curve.setter
    def min_nodes_curve(self, value: float):
        gmsh.option.set_number("Mesh.MinimumCurveNodes", value)



from __future__ import print_function

import os
import compas


__author__ = ["tom van mele"]
__copyright__ = "Block Research Group - ETH Zurich"
__license__ = "MIT License"
__email__ = "van.mele@arch.ethz.ch"
__version__ = "0.4.4"


HERE = os.path.dirname(__file__)

HOME = os.path.abspath(os.path.join(HERE, "../../"))
DATA = os.path.abspath(os.path.join(HOME, "data"))
DOCS = os.path.abspath(os.path.join(HOME, "docs"))
TEMP = os.path.abspath(os.path.join(HOME, "temp"))

# Check if package is installed from git
# If that's the case, try to append the current head's hash to __version__
try:
    git_head_file = compas._os.absjoin(HOME, ".git", "HEAD")

    if os.path.exists(git_head_file):
        # git head file contains one line that looks like this:
        # ref: refs/heads/master
        with open(git_head_file, "r") as git_head:
            _, ref_path = git_head.read().strip().split(" ")
            ref_path = ref_path.split("/")

            git_head_refs_file = compas._os.absjoin(HOME, ".git", *ref_path)

        if os.path.exists(git_head_refs_file):
            with open(git_head_refs_file, "r") as git_head_ref:
                git_commit = git_head_ref.read().strip()
                __version__ += "-" + git_commit[:8]

except Exception:
    pass

__all__ = ["HOME", "DATA", "DOCS", "TEMP"]
